# Author Darrell Best: bbest@g.clemson.edu
# Date 10/06/2014
#
# Singleton State class holds global variables

import sys
import time
from collections import deque

from rstats import RunningStats

# prevent Python from writing .pyc files
sys.dont_write_bytecode = True

class State:
  
    def __init__(self):
        self.mouse_coord = [.5, .5]
        self.gaze_coord = [.5, .5]
        self.fullscreen = 0
        self.butterworth = 0
        self.input = 0 # 1 is eye tracker, 0 is mouse
        self.dfwidth = 10
        self.dfdegree = 2
        self.dfo = 1
        self.threshold = 20
        self.fixating = 0
        self.st = 0.0		# current fixation start time
        self.et = 0.0		# current fixation end time
        self.tt = 0.0		# current fixation total time (duration)
        self.ux = 0.0		# current fixation mean x coord
        self.uy = 0.0		# current fixation mean y coord
        self.sx = 0.0		# current fixation x coord stdev
        self.sy = 0.0		# current fixation y coord stdev
        self.k = 0.0		# current fixation count
        self.uamp = 0.0		# current mean inter-fixation amplitude
        self.samp = 0.0		# current mean -fixation stdev
        self.udur = 0.0		# current mean fixation duration
        self.sdur = 0.0		# current mean fixation stdev
        self.K = 0.0		# current K coefficient
        self.rx = RunningStats()# current running statistics for x coord
        self.ry = RunningStats()# current running statistics for y coord
        self.points = deque(maxlen=30)
        self.fixations = deque(maxlen=30)
        self.vms = 0.0
        self.vtd = 0.0

    def set_input(self, flag):
        self.input = flag

    def set_mouse_coord_x(self, x):
        self.mouse_coord[0] = x

    def set_mouse_coord_y(self, y):
        self.mouse_coord[1] = y

    def set_gaze_coord_x(self, x):
        self.gaze_coord[0] = x

    def set_gaze_coord_y(self, y):
        self.gaze_coord[1] = y

    def toggle_butterworth(self):
        self.butterworth = 1 - self.butterworth

    def set_butterworth(self, flag):
        self.butterworth = flag

    def set_dfwidth(self,d):
        self.dfwidth = d

    def set_dfdegree(self,d):
        self.dfdegree = d

    def set_dfo(self,d):
        self.dfo = d

    def set_threshold(self,T):
        self.threshold = T

    def set_fixating(self,f):
        self.fixating = f

    def get_input(self):
        return self.input

    def get_mouse_coord_x(self):
        return self.mouse_coord[0]
    
    def get_mouse_coord_y(self):
        return self.mouse_coord[1]
    
    def get_gaze_coord_x(self):
        return self.gaze_coord[0]
    
    def get_gaze_coord_y(self):
        return self.gaze_coord[1]
    
    def get_butterworth(self):
        return self.butterworth
    
    def get_dfwidth(self):
        return self.dfwidth

    def get_dfdegree(self):
        return self.dfdegree

    def get_dfo(self):
        return self.dfo

    def get_threshold(self):
        return self.threshold

    def get_fixating(self):
        return self.fixating

    def set_fullscreen(self,flag):
        self.fullscreen = flag

    def get_fullscreen(self):
        return self.fullscreen

    def toggle_fullscreen(self):
        self.fullscreen = 1 - self.fullscreen
